#include "xlistobject.h"
#include "xiterobject.h"
#include "kernel.h"
#include "xquery.h"


#define DEFERRED_ADDRESS(ADDR) 0

typedef int (*cmp_fun_t)(PyObject*, PyObject*);

static PyObject* uri;

static PyXListInfo* xlist_tpinfo;

static PyObject* __tag__;
static PyObject* __uri__;
static PyObject* __tpinfo__;
static PyObject* __dict__;
static PyObject* __xlist_atv__;
static PyObject* __xlist_itv__;
static PyObject* __xlist_required__;


staticforward PyMethodDef xlist_methods[];


PyObject* 
Attr_FullName(PyObject* prefix, PyObject* tag) {
	PyObject* fullname;
	
	if (prefix == Py_None) {
		Py_INCREF(tag);
		return tag;
	}

	Py_INCREF(prefix);
	fullname = prefix;
	PyString_ConcatAndDel(&fullname, PyString_InternFromString(":"));
	PyString_Concat(&fullname, tag);

	return fullname;
}

int
Attr_SplitFullName(PyObject* fullname, PyObject** prefix, PyObject** tag) {
	char* fullname_c;
	int sep = -1;
	int i;
	
	fullname_c = PyString_AsString(fullname);
	for (i=0; fullname_c[i]; ++i) 
		if (fullname_c[i] == ':')
			sep = i;

	if (sep < 0) {
		Py_INCREF(Py_None);
		*prefix = Py_None;
		Py_INCREF(fullname);
		*tag = fullname;
	}
	else {
		*prefix = PyString_FromStringAndSize(fullname_c, sep);
		*tag = PyString_FromString(fullname_c+sep+1);
	}
	return 0;
}

PyObject* PyString_FromExc(PyObject* key) {
	PyObject* ptype;
	PyObject* pvalue;
	PyObject* ptraceback;
	PyObject* error;

	PyErr_Fetch(&ptype, &pvalue, &ptraceback);
	if (pvalue == NULL || pvalue == Py_None) 
		if (ptype == NULL)
			error = PyString_FromFormat("Attribute %s not valid",
				PyString_AsString(key));
		else
			error = PyString_FromFormat("Attribute %s not valid: %s ",
				PyString_AsString(key), 
				PyString_AsString(PyObject_Repr(ptype)));
	else
		error = PyString_FromFormat("Attribute %s not valid: %s (%s)",
			PyString_AsString(key), 
			PyString_AsString(PyObject_Repr(ptype)),
			PyString_AsString(PyObject_Repr(pvalue)));
	return error;
}

PyObject* 
Attr_AsXML(PyObject* key, PyObject* value, PyObject* datatype) {
	PyObject* method;
	
	if (datatype == Py_None) {
		Py_INCREF(value);
		return value;
	}

	method = PyObject_GetAttrString(datatype, "py2xml");
	value = PyObject_CallFunctionObjArgs(method, value, NULL);
	return value;
}

PyObject* 
Attr_AsPY(PyObject* key, PyObject* value, PyObject* datatype) {
	PyObject* method;
	
	if (datatype == Py_None) {
		Py_INCREF(value);
		return value;
	}

	method = PyObject_GetAttrString(datatype, "xml2py");
	value = PyObject_CallFunctionObjArgs(method, value, NULL);
	return value;
}

PyObject*
PyXList_New() {
	PyXListObject* xlist;
	xlist = (PyXListObject*) PyObject_GC_New(PyXListObject, &PyXList_Type);
	if (xlist == NULL) {
		return NULL;
	}
	xlist->list.ob_item = NULL;
	xlist->list.ob_size = 0;
	xlist->attrs = PyDict_New();
	xlist->tpinfo = xlist_tpinfo;
	
	PyObject_GC_Track(xlist);

	return (PyObject*) xlist;
}

PyObject*
PyXList_PoorCopy(PyObject* self) {
	PyXListObject* xlist;
	PyXListObject* tself;
	
	tself = (PyXListObject*) self;
	xlist = (PyXListObject*) PyObject_GC_New(PyXListObject, &PyXList_Type);
	if (xlist == NULL) {
		return NULL;
	}
	xlist->list.ob_item = NULL;
	xlist->list.ob_size = 0;

	xlist->attrs = PyDict_Copy(tself->attrs);
	xlist->tpinfo = tself->tpinfo;

	return (PyObject*)xlist;
}


PyObject*
PyXList_Copy(PyObject* self) {
	PyXListObject* xlist;
	PyXListObject* tself;
	PyObject* item;
	int i, size;
	
	tself = (PyXListObject*) self;
	xlist = (PyXListObject*) PyObject_GC_New(PyXListObject, &PyXList_Type);
	if (xlist == NULL) {
		return NULL;
	}
	xlist->list.ob_item = NULL;
	xlist->list.ob_size = 0;

	xlist->attrs = PyDict_Copy(tself->attrs);
	xlist->tpinfo = tself->tpinfo;

	size = PyList_GET_SIZE(tself);
	for (i=0; i < size; ++i) {
		item = PyList_GET_ITEM(tself, i);
		if (PyXList_Check(item)) {
			item = PyXList_Copy(item);
			if (item == NULL) {
				Py_DECREF((PyObject*)xlist);
				return NULL;
			}
			PyList_Append((PyObject*)xlist, item);
			Py_DECREF(item);
		}
		else
			PyList_Append((PyObject*)xlist, item);

	}
	return (PyObject*)xlist;
}


PyObject* kernel_strip_c(PyObject *seq);


static PyObject*
xlist_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
	PyXListObject* self;
	PyXListInfo* tpinfo;


	Satine_WriteMoreInfo("xlist_new: type %s\n", type->tp_name);
	
	tpinfo = PyXListInfo_FromType(type);
	if (tpinfo == NULL) 
		return NULL;

	self = (PyXListObject*) PyList_Type.tp_new(type, args, kwds);
	if (!self)
		return NULL;

	self->tpinfo = tpinfo;
	self->attrs = PyDict_New();

	Satine_WriteMoreInfo("xlist_new: end\n");
	return (PyObject*) self;
}
/*
static PyObject*
xattr_from_fullname_and_py(PyObject* fullname, PyObject* py) {
	PyObject* prefix;
	PyObject* key;
	PyObject* uri;
	PyObject* xattr;

	Attr_SplitFullName(fullname, &prefix, &key);
	if (prefix == Py_None) {
		xattr = PyXAttr_FromPY(prefix, key, Py_None, py);
		Py_DECREF(prefix);
		Py_DECREF(key);
	}
	else {
		uri = PyDict_GetItem(default_uris, prefix);
		if (uri == NULL) {
			PyErr_SetObject(PyExc_AttributeError, PyString_FromFormat("undefined prefix '%s'", 
				PyString_AsString(prefix)));
			Py_DECREF(key);
			Py_DECREF(prefix);
			return NULL;
		}
		xattr = PyXAttr_FromPY(prefix, key, uri, py);
		Py_DECREF(prefix);
		Py_DECREF(key);
	}
	return xattr;
}
*/
static int
xlist_init(PyXListObject *self, PyObject *args, PyObject *kwds)
{
	int pos;
	int size;
	PyObject* key;
	PyObject* w_args;
	PyObject* none_attributes;	
	
	Satine_WriteMoreInfo("xlist_init: type %s\n", self->list.ob_type->tp_name);

	w_args = PyTuple_New(1);
	none_attributes = self->tpinfo->none_attrs;	

	Py_INCREF(args);
	PyTuple_SetItem(w_args, 0, args);
	if (PyList_Type.tp_init((PyObject *)self, w_args, NULL) < 0) {
		return -1;
	}

	size = PyList_GET_SIZE(none_attributes);
	for (pos=0; pos < size; ++pos) {
		key = PyList_GET_ITEM(none_attributes, pos);
		PyDict_SetItem(self->attrs, key, Py_None);
	}

	if (kwds) 
		PyDict_Merge(self->attrs, kwds, 1);

	PyErr_Clear();
	Py_DECREF(w_args);
	Satine_WriteMoreInfo("xlist_init: end\n");

	return 0;
}

static void
xlist_dealloc(PyXListObject *op)
{
	Py_DECREF(op->attrs);

	PyList_Type.tp_dealloc((PyObject*)op);
}

static PyObject *
xlist_getattro(PyXListObject *self, PyObject *name)
{
	PyObject* attr;
	
	attr = PyDict_GetItem(self->attrs,name);
	if (attr) {
		PyErr_Clear();
		Py_INCREF(attr);
		return attr;
	}

	attr = PyObject_GenericGetAttr((PyObject*)self, name);
	if (attr) {
		PyErr_Clear();
		return attr;
	}

	if (PyObject_Compare(name, __dict__) == 0) {
		PyErr_Clear();
		Py_INCREF(self->attrs);
		return self->attrs;
	}	

	PyErr_Clear();
	PyErr_SetString(PyExc_AttributeError, "Attribute not available");
	return NULL;
}

int
xlist_setattro(PyXListObject *self, PyObject* name, PyObject* value) {
	return PyDict_SetItem(self->attrs, name, value);
}


static void 
xlist_repr_ns(PyObject* uri, PyObject* prefix, PyObject** str, PyObject* declared_ns, int strict) {
	int emptyprefix;
	char* uri_c;
	emptyprefix = PyString_Size(prefix) == 0;

	if (!strict)
		return;

	if (!PySequence_Contains(declared_ns, uri)) {
		uri_c = PyString_AsString(uri);
		if (strncmp(uri_c, "http://satine.sourceforge.net/temp/dyn/", 
				strlen("http://satine.sourceforge.net/temp/dyn/")) != 0) {
			PyString_ConcatAndDel(str, PyString_FromString(" xmlns"));
			if (!emptyprefix) {
				PyString_ConcatAndDel(str, PyString_FromString(":"));
				PyString_Concat(str, prefix);
			}
			PyString_ConcatAndDel(str, PyString_FromString("=\""));
			PyString_Concat(str, uri);
			PyString_ConcatAndDel(str, PyString_FromString("\""));
		}
		PyList_Append(declared_ns, uri);
	}
}

PyObject*
xlist_xml_str(PyObject* value, int strict) {
	char* cstr;
	char* nstr;
	int i, j;
	int len;
	PyObject* str;
	
	str = PyObject_Str(value);
	if (!strict)
		return str;

	PyString_AsStringAndSize(str, &cstr, &len);
	nstr = malloc(len*7);

	for (i=0, j=0; cstr[i]; ++i) {
		if (cstr[i] == '&') {
			strcpy(nstr+j, "&amp;");
			j+=5;
		}
		else if (cstr[i] == '<') {
			strcpy(nstr+j, "&lt;");
			j+=4;
		}
		else if (cstr[i] == '>') {
			strcpy(nstr+j, "&gt;");
			j+=4;
		}
		else if (cstr[i] == '"') {
			strcpy(nstr+j, "&quote;");
			j+=7;
		}
		else 
			nstr[j++] = cstr[i];
	}
	nstr[j] = 0;
	Py_DECREF(str);
	str = PyString_FromString(nstr);
	free(nstr);
	return str;
}

static PyObject *
xlist_attr_keyrepr(PyObject* prefix, PyObject* tag, PyObject* suggested_prefixes) {
	PyObject* sug_prefix = NULL;
	PyObject* uri;
	PyObject* res;


	if (prefix == Py_None) {
		Py_INCREF(tag);
		res = tag;
	}
	else {
		uri = PyDict_GetItem(default_uris, prefix);
		if (uri) 
			sug_prefix = PyDict_GetItem(suggested_prefixes, uri);
		if (sug_prefix)
			res = Attr_FullName(sug_prefix, tag);
		else
			res = Attr_FullName(prefix, tag);

	}
	return res;
}

static PyObject*
xlist_xattr_valrepr(PyObject* key, PyObject* value, int strict, PyXListInfo* tpinfo) {
	PyObject* datatype;
	if (!strict) {
		if (PyString_CheckExact(value)) {
			Py_INCREF(value);
			return value;
		}
		else
			return PyObject_Repr(value);
	}
	
	datatype = PyDict_GetItem(tpinfo->attrs_types, key);
	if (datatype) {
		value = Attr_AsXML(key, value, datatype);
		return value;
	}
	else {
		if (PyString_CheckExact(value)) {
			Py_INCREF(value);
			return value;
		}
		else
			return PyObject_Repr(value);
	}
}

PyObject *
xlist_repr_c(PyXListObject *v, PyObject* suggested_prefixes, PyObject* declared_ns, int strict)
{
	int i;
	int size;
	int pos;
	int emptyprefix;
	PyObject *s;
	PyObject *key, *value;
	PyObject *aprefix;
	PyObject *atag;
	PyObject *auri;
	PyObject *avalue;
	PyObject *prefix;
	PyObject* uri;
	PyObject* tag;
	PyObject* fullname;

	PyObject *pieces = NULL, *result = NULL;
	PyObject* quote = PyString_InternFromString("\"");
	PyObject* space = PyString_InternFromString(" ");
	PyObject* equal = PyString_InternFromString("=");
	PyObject* none = PyString_InternFromString("");
	PyObject* cr = PyString_InternFromString("\n");

	Satine_WriteMoreInfo("xlist repr\n");
	
	i = Py_ReprEnter((PyObject*)v);
	if (i != 0) {
		return i > 0 ? PyString_FromString("<.../>") : NULL;
	}

	size = PyList_Size((PyObject*)v);

	s = PyString_FromString("<");

	uri = PyXList_GetUri(v);
	tag = PyXList_GetTag(v);
	prefix = PyDict_GetItem(suggested_prefixes, uri);
	if (!prefix)
		prefix = PyXList_GetPrefix(v);
	emptyprefix = PyString_Size(prefix) == 0;

	PyString_Concat(&s, prefix);
	if (!emptyprefix) 
		PyString_ConcatAndDel(&s, PyString_FromString(":"));
	PyString_Concat(&s, tag);

	xlist_repr_ns(uri, prefix, &s, declared_ns, strict);

	pos = 0;
	while (PyDict_Next(v->attrs, &pos, &key, &value)) {
		if (value == Py_None)
			continue;

		Satine_WriteMoreInfo("1\n");
		Attr_SplitFullName(key, &aprefix, &atag);
		fullname = xlist_attr_keyrepr(aprefix, atag, suggested_prefixes);
		if (fullname == NULL)
			return NULL;

		if (aprefix != Py_None) {
			auri = PyDict_GetItem(default_uris, aprefix);
			if (auri)
				xlist_repr_ns(auri, aprefix, &s, declared_ns, strict);
		}
		Py_DECREF(atag);
		Py_DECREF(aprefix);

		PyString_Concat(&s, space);
		PyString_ConcatAndDel(&s, fullname);
		PyString_Concat(&s, equal);
		PyString_Concat(&s, quote);


		avalue = xlist_xattr_valrepr(key, value, strict, v->tpinfo);
		if (!avalue)
			return NULL;


		PyString_ConcatAndDel(&s, avalue);
		PyString_Concat(&s, quote);
	}

	result = s;

	if (size == 0) {
		PyString_ConcatAndDel(&result, PyString_FromString("/>"));
		Py_ReprLeave((PyObject*)v);
		goto Done;
	}

	PyString_ConcatAndDel(&result, PyString_FromString(">\n"));

	pieces = PyList_New(0);

	/* Do repr() on each element.  Note that this may mutate the list,
	   so must refetch the list size on each iteration. */
	for (i = 0; i < size; ++i) {
		int status;
		value = PyList_GetItem((PyObject*)v,i);
		if PyXList_Check(value) {
			s = xlist_repr_c((PyXListObject*)value, suggested_prefixes, declared_ns, strict);
		}
		else  
			s = xlist_xml_str(value, strict);	

		if (s == NULL)
			goto Done;
		status = PyList_Append(pieces, s);
		Py_DECREF(s);  /* append created a new ref */
		if (status < 0)
			goto Done;
	}

	/* Paste them all together with ", " between. */
	s = _PyString_Join(cr, pieces);
	PyString_ConcatAndDel(&result, s);

	s = PyString_FromString("\n</");

	PyString_Concat(&s, prefix);
	if (PyString_Size(prefix)) 
		PyString_ConcatAndDel(&s, PyString_FromString(":"));
	PyString_Concat(&s, tag);

	PyString_ConcatAndDel(&s, PyString_FromString(">"));
	PyString_ConcatAndDel(&result, s);

Done:
	Py_XDECREF(pieces);
	Py_ReprLeave((PyObject *)v);
	Py_DECREF(space);
	Py_DECREF(equal);
	Py_DECREF(none);
	Py_DECREF(cr);
	Py_ReprLeave((PyObject*)v);
	Satine_WriteMoreInfo("xlist repr end\n");
	return result;
}

static PyObject*
xlist_possible_items(PyObject* node) {
	int nsize;
	int i;
	PyObject* item;
	PyObject* final;
	PyObject* res;
	PyObject* label;

	res = PyList_New(0);
	nsize = PyList_GET_SIZE(node);
	for (i=0; i < nsize; ++i) {
		item = PyList_GET_ITEM(node, i); 
		label = PyTuple_GET_ITEM(item, 0);
		PyList_Append(res, label);
	}
	
	final = PyObject_GetAttrString(node, "final");
	if (PyInt_AsLong(final)) {
		Py_INCREF(Py_None);
		PyList_Append(res, Py_None);
	}
	return res;
}

static int
xlist_validate_attrs(PyXListObject* self, PyObject* res) {
	int pos;
	int rsize;
	int i;
	PyObject* item;
	PyObject* key;
	PyObject* value;
	PyObject* dtcls;
	PyObject* newvalue;

	PyObject* error;
	PyObject* default_attributes = self->tpinfo->default_attrs;
	pos = 0;

	if (self->tpinfo->attrs_types == Py_None)
		return 1;

	while (PyDict_Next(self->attrs, &pos, &key, &value)) {
		if (value == Py_None) {
			value = PyDict_GetItem(default_attributes, key);
			if (value == NULL)
				continue;
			else {
				PyDict_SetItem(self->attrs, key, value);
			}
		}

		dtcls = PyDict_GetItem(self->tpinfo->attrs_types, key);
		if (!dtcls) {
			error = PyString_FromFormat("Attribute %s is not in the element schema",
				PyString_AsString(PyObject_Repr(key)));
			PyList_Append(res, error);
			return 0;
		}

		newvalue = Attr_AsPY(key, value, dtcls);
		if (!newvalue) {
			error = PyString_FromExc(key);
			PyList_Append(res, error);
			return 0;
		}
		else {
			PyDict_SetItem(self->attrs, key, newvalue);
		}
	}

	rsize = PySequence_Size(self->tpinfo->required_attrs);
	for (i=0; i < rsize; ++i) {
		key = PyList_GetItem(self->tpinfo->required_attrs, i);
		item = PyDict_GetItem(self->attrs, key);
		if (item == NULL) {
			error = PyString_FromFormat("Attribute %s is required",
				PyString_AsString(key));
			PyList_Append(res, error);
			return 0;
		}
	}
	return 1;
}


static PyObject*
xlist_next_item(PyXListObject* self) {
	int esize;
	int nsize;
	int i;
	PyObject* item;
	PyObject* node;
	PyObject* final;
	PyObject* res;
	PyObject* label;
	PyObject* newnode;
	PyObject* needed;
	PyObject* error_str;
	char* needed_c;
	char* item_c;

	res = PyList_New(0);
	node = self->tpinfo->items_dfa;
	if (node != Py_None) {
		esize = PyList_GET_SIZE(self);
		for (i=0; i < esize; ++i) {
			item = PyList_GET_ITEM(self, i); 
			node = xquery_next(item, node);
			newnode = xquery_next(item, node);
			if (!newnode) {
				needed = PyObject_Repr(xlist_possible_items(node));
				needed_c = PyString_AsString(needed);
				item_c = PyString_AsString(PyObject_Repr(item));
				error_str = PyString_FromFormat("error at position %i: found %s while one of %s needed",
					i, item_c, needed_c);
				PyList_Append(res, error_str);
				PyList_Insert(res, 0, PyInt_FromLong(i));
				return NULL;
			}
			node = newnode;
		}

		nsize = PyList_GET_SIZE(node);
		for (i=0; i < nsize; ++i) {
			item = PyList_GET_ITEM(node, i); 
			label = PyTuple_GET_ITEM(item, 0);
			PyList_Append(res, label);
		}
		
		final = PyObject_GetAttrString(node, "final");
		if (PyInt_AsLong(final)) {
			Py_INCREF(Py_None);
			PyList_Append(res, Py_None);
		}
	}
	return res;
}


static int
xlist_validate_items(PyXListObject* self, PyObject* res, PyObject* alternatives) {
	int esize;
	int i;
	int child_valid;
	PyObject* item;
	PyObject* node;
	PyObject* newnode;
	PyObject* final;
	PyObject* needed;
	PyObject* error_str;
	char* needed_c;
	char* item_c;

	if (!xlist_validate_attrs(self, res))
		return 0;
		
	esize = PyList_GET_SIZE(self);
	for (i=0; i < esize; ++i) {
		item = PyList_GET_ITEM(self, i);
		if (PyXList_Check(item)) {
			child_valid = xlist_validate_items((PyXListObject*)item, res, alternatives);
			if (!child_valid) {
				PyList_Insert(res, 0, PyInt_FromLong(i));
				return 0;
			}
		}

	}

	node = self->tpinfo->items_dfa;
	if (node != Py_None) {
		for (i=0; i < esize; ++i) {
			item = PyList_GET_ITEM(self, i); 
			newnode = xquery_next(item, node);
			if (!newnode) {
				needed = PyObject_Repr(xlist_possible_items(node));
				needed_c = PyString_AsString(needed);
				item_c = PyString_AsString(PyObject_Repr(item));
				error_str = PyString_FromFormat("error at position %i: found %s while one of %s needed",
					i, item_c, needed_c);
				PyList_Append(res, error_str);
				PyList_Insert(res, 0, PyInt_FromLong(i));
				return 0;
			}
			node = newnode;
		}
		final = PyObject_GetAttrString(node, "final");
		if (!PyInt_AsLong(final)) {
			PyList_Append(res, PyString_FromString("not enough items"));
			PyList_Insert(res, 0, PyInt_FromLong(i));
			return 0;
		}
	}
	return 1;
}

static PyObject*
xlist_validate(PyXListObject* self) {
	PyObject* res = PyList_New(0);
	PyObject* alternatives = PyList_New(0);

	Satine_WriteMoreInfo("xlist validation\n");

	xlist_validate_items(self, res, alternatives);
	Py_DECREF(alternatives);
	return res;
}

static PyObject *
xlist_repr(PyXListObject *v) {
	PyObject* suggested_prefixes = PyDict_New();
	PyObject* declared_ns = PyList_New(0);
	PyObject* res;	

	res = xlist_repr_c(v, suggested_prefixes, declared_ns, py2xml_strict);
	Py_DECREF(suggested_prefixes);
	Py_DECREF(declared_ns);
	return res;
}

static int
xlist_print(PyXListObject* op, FILE *fp, int flags)
{
	char* data;
	PyObject* res;
	
	res = xlist_repr(op);
	if (res == NULL) {
		return -1;
	}

	data = PyString_AsString(res);
	fprintf(fp, data);
	fflush(fp);
	
	Py_DECREF(res);
	return 0;
	
}


int
xlist_match_c(PyXListObject * list, PyXListObject * pattern) {
	int pos = 0;
	PyObject* key;
	PyObject* value;
	PyObject* s_value;
	PyObject* cmp;
	
	if (PyObject_Compare(PyXList_GetTag(list), PyXList_GetTag(pattern)) != 0)
		return 0;

	if (PyObject_Compare(PyXList_GetUri(list), PyXList_GetUri(pattern)) != 0)
		return 0;

	while (PyDict_Next(pattern->attrs, &pos, &key, &value)) {
		if (value != Py_None) {
			s_value = PyDict_GetItem(list->attrs, key);
			if (s_value) {
				if (PyFunction_Check(value)) {
					cmp = PyObject_CallFunction(value, "O", s_value);
					if (!PyInt_AsLong(cmp))
						return 0;
				}
				else 
					if (PyObject_Compare(s_value, value) != 0)
						return 0;
			}
			else
				return 0;
		}
	}

	return 1;
}

static PyObject*
xlist_match(PyXListObject * list, PyObject * pattern) {
	if (!PyXList_Check(pattern)) {
		PyErr_SetString(PyExc_TypeError, "Argument must be a xlist");
		return NULL;
	}
	return PyInt_FromLong(xlist_match_c(list, (PyXListObject*)pattern));
}



static PyObject *
xlist_getiter(PyObject *self)
{
	return PyXIter_New_c(self, NULL, "");
}

static PyObject *
xlist_iter(PyObject *self, PyObject* args, PyObject* kwds) {
    PyObject* criteria = NULL;
    char *type = "xre";
	int start = 0;
	int end = -1;

    static char *kwlist[] = {"pattern", "style", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|Os", kwlist, 
                                     &criteria, &type))
        return NULL; 

	return PyXIter_New_c(self, criteria, type);
}

PyObject*
xlist_query(PyObject* self, PyObject *args, PyObject* kwds) {
	PyObject* ptr;
	PyObject* iterator;
	PyObject* res = PyXList_New();

	Satine_WriteMoreInfo("xlist query\n");

	iterator = xlist_iter(self, args, kwds);
	if (iterator == NULL) {
		PyErr_SetString(PyExc_AttributeError, "bad parameters in query");
		return NULL;
	}

	for (ptr = PyIter_Next(iterator); ptr ;ptr = PyIter_Next(iterator)) 
		PyList_Append(res, ptr);			

	Py_DECREF(iterator);
	return res;
}

PyObject*
xlist_visit(PyObject* self, PyObject *args) {
	PyObject* items;

	Satine_WriteMoreInfo("xlist visit\n");

    if (!PyArg_ParseTuple(args, "O!", &PyList_Type, &items))
		return NULL;

	return xquery_visit(self, items);
}


#define xlist_copy PyXList_Copy


static int
xlist_length(PyXListObject *list)
{
	return PyList_GET_SIZE(list);
}

static PyObject *
xlist_subscript(PyXListObject *list, register PyObject *key)
{
	int idx;
	PyObject* res;

	if (PyInt_CheckExact(key)) {
		idx = PyInt_AsLong(key);
		res = PyList_GetItem((PyObject*)list, idx);
		if (res)
			Py_INCREF(res);
		return res;
	}
	else {
		res = PyDict_GetItem(list->attrs, key);
		if (res) {
			Py_INCREF(res);
			return res;
		}
		else {
			PyErr_SetObject(PyExc_KeyError, key);
			return NULL;
		}
	}
}

static int
xlist_ass_sub(PyXListObject *list, PyObject *key, PyObject *w)
{
	int idx;
	PySequenceMethods *m;
	PyObject* prefix;
	PyObject* tag;
	PyObject* uri;

	m = PyList_Type.tp_as_sequence;

	if (PyInt_CheckExact(key)) {
		idx = PyInt_AsLong(key);
		return m->sq_ass_item((PyObject*)list, idx, w);
	}
	else
		if (w == NULL) 
			return PyDict_DelItem(list->attrs, key);
		else {
			Attr_SplitFullName(key, &prefix, &tag);
			if (prefix == Py_None) {
				Py_DECREF(prefix);
				Py_DECREF(tag);
				return PyDict_SetItem(list->attrs, key, w);
			}
			else {
				uri = PyDict_GetItem(default_uris, prefix);
				if (!uri) {
					PyErr_SetObject(PyExc_KeyError, PyString_FromFormat("undefined prefix '%s'", 
						PyString_AsString(prefix)));
					Py_DECREF(tag);
					Py_DECREF(prefix);
					return -1;
				}
				Py_DECREF(prefix);
				Py_DECREF(tag);
				return PyDict_SetItem(list->attrs, key, w);
			}
		}
}

static int
xlist_traverse(PyListObject *o, visitproc visit, void *arg)
{
	int i, err;
	PyObject *x;

	for (i = o->ob_size; --i >= 0; ) {
		x = o->ob_item[i];
		if (x != NULL) {
			err = visit(x, arg);
			if (err)
				return err;
		}
	}
	return 0;
}

static int
xlist_clear(PyXListObject *lp)
{
	(void) PyList_SetSlice((PyObject *)lp, 0, lp->list.ob_size, 0);
	(void) PyDict_Clear((PyObject *)lp->attrs);
	return 0;
}

static int
xlist_eqcmp(PyXListObject *self, PyObject *w) {
	PyObject* tmp;
	int lcmp, dcmp, ocmp;
	PyTypeObject* list_type = (PyTypeObject*) &PyDict_Type;
	PyXListObject* wx;
	if (!PyXList_Check(w)) 
		return 0;

	wx = (PyXListObject*) w;
	ocmp = PyObject_Compare(self->tpinfo->tag, wx->tpinfo->tag) | 
		PyObject_Compare(self->tpinfo->uri, wx->tpinfo->uri);
	if (ocmp != 0)
		return 0;


	tmp = PyList_Type.tp_richcompare((PyObject*)self, w, Py_EQ);
	if (!tmp)
		return -1;
	lcmp = PyInt_AS_LONG(tmp);
	if (lcmp != 1)
		return lcmp;

	tmp = PyDict_Type.tp_richcompare(self->attrs, wx->attrs, Py_EQ);
	dcmp = PyInt_AS_LONG(tmp);
	return dcmp;
}

static int
xlist_necmp(PyXListObject *self, PyObject *w) {
	int eq;

	eq = xlist_eqcmp(self, w);
	switch(eq) {
		case 0: return 1;
		case 1: return 0;
		case -1: return -1;
	}
	return -1;
}

static int
xlist_ltcmp(PyXListObject *self, PyObject *w) {
	int dcmp, ocmp, err;
	PyXListObject* wx;
	if (!PyXList_Check(w)) 
		if (PyInt_Check(w))
			return 0;
		else
			return 1;

	wx = (PyXListObject*) w;
	err = PyObject_Cmp(self->tpinfo->uri, wx->tpinfo->uri, &ocmp);
	if (err == -1)
		return -1;
	if (ocmp == -1)
		return 1;
	else if (ocmp == 1)
		return 0;

	err = PyObject_Cmp(self->tpinfo->tag, wx->tpinfo->tag, &ocmp);
	if (err == -1)
		return -1;
	if (ocmp == -1)
		return 1;
	else if (ocmp == 1)
		return 0;

	err = PyObject_Cmp(self->attrs, wx->attrs, &dcmp);
	if (err == -1)
		return -1;
	if (dcmp == -1)
		return 1;
	else
		return 0;

	return dcmp;
}

static int
xlist_gtcmp(PyXListObject *self, PyObject *w) {
	int dcmp, ocmp, err;
	PyXListObject* wx;
	if (!PyXList_Check(w)) 
		if (PyInt_Check(w))
			return 1;
		else
			return 0;

	wx = (PyXListObject*) w;
	err = PyObject_Cmp(self->tpinfo->uri, wx->tpinfo->uri, &ocmp);
	if (err == -1)
		return -1;
	if (ocmp == 1)
		return 1;
	else if (ocmp == 1)
		return 0;

	err = PyObject_Cmp(self->tpinfo->tag, wx->tpinfo->tag, &ocmp);
	if (err == -1)
		return -1;
	if (ocmp == 1)
		return 1;
	else if (ocmp == 1)
		return 0;

	err = PyObject_Cmp(self->attrs, wx->attrs, &dcmp);
	if (err == -1)
		return -1;
	if (dcmp == 1)
		return 1;
	else
		return 0;

	return dcmp;
}

static int 
xlist_lecmp(PyXListObject *self, PyObject *w) {
	int gt;

	gt = xlist_gtcmp(self, w);
	switch(gt) {
		case 0: return 1;
		case 1: return 0;
		case -1: return -1;
	}
	return -1;
}

static int 
xlist_gecmp(PyXListObject *self, PyObject *w) {
	int lt;

	lt = xlist_gtcmp(self, w);
	switch(lt) {
		case 0: return 1;
		case 1: return 0;
		case -1: return -1;
	}
	return -1;
}

static PyObject *
xlist_richcompare(PyXListObject *self, PyObject *w, int op)
{
	int res;
	switch(op) {
		case Py_LT: res = xlist_ltcmp(self, w); break;
		case Py_LE: res = xlist_lecmp(self, w); break;
		case Py_EQ: res = xlist_eqcmp(self, w); break;
		case Py_NE: res = xlist_necmp(self, w); break;
		case Py_GT: res = xlist_gtcmp(self, w); break;
		case Py_GE: res = xlist_gecmp(self, w); break;
		default: PyErr_SetString(PyExc_NotImplementedError, "unknown operator during richcompare");
				 return NULL;
	}

	switch(res) {
		case -1: return NULL;
		case 0: return PyInt_FromLong(0);
		case 1: return PyInt_FromLong(1);
	}
	return NULL;
}


static PyMappingMethods xlist_as_mapping = {
	(inquiry)xlist_length, /*mp_length*/
	(binaryfunc)xlist_subscript, /*mp_subscript*/
	(objobjargproc)xlist_ass_sub, /*mp_ass_subscript*/
};



static char iter_doc[] = 
"L.iter(iter([criteria[, style]])->iterator\n\n\
Returns an iterator that satisfies the criteria. The parameter style points out\n\
what format is criteria. When style is missing, criteria has the syntax\n\
\n\
\tcriteria := 'xre|[attr1[, attr2[,..attrN]]]'\n\
\n\
xre is a XRE expression that identifies matching items; attr1..attrN are the\n\
attributes that the iterator extracts from the matching item.\n\
If style=='tag', criteria is the tag of matching items.\n\
If style=='pyfun', criteria is a python function tha decides if the item belongs\n\
to the matching set.";

static char query_doc[] = 
"L.query(iter([criteria[, style]])->xlist\n\n\
Returns all items that satisfy the criteria. The parameter style points out\n\
what format is criteria. When style is missing, criteria has the syntax\n\
\n\
\tcriteria := 'xre|[attr1[, attr2[,..attrN]]]'\n\
\n\
xre is a XRE expression that identifies matching items; attr1..attrN are the\n\
attributes that the iterator extracts from the matching item.\n\
If style=='tag', criteria is the tag of matching items.\n\
If style=='pyfun', criteria is a python function tha decides if the item belongs\n\
to the matching set.";

static char match_doc[] = 
"L.match(obj)\n\n\
Check if obj has uri, tag and attributes like L. The matching regards only attributes\n\
that are present in both objects.";

static char validate_doc[] = 
"L.validate(obj) -> list\n\n\
Check if the xlist is compliant with its __attrs__ and __items__ fields. If the xlist is\n\
compliant, the result is an empty list. Otherwise the result has the format:\n\
\n\
\t[pos0[,pos1[,pos2...posN]], error]\
\n\
where pos0..posN are the positions in the xlist structure where the validation fails and\n\
error is the reason. Also the method converts all attributes to the type according to the\n\
information in __attrs__.";

static char next_item_doc[] = 
"L.next_item() -> list\n\n\
Return a list of objects that can be valid items for L, according to the fiels __items__\n\
and the actual content.";

static char copy_doc[] = 
"L.copy() -> xlist\n\n\
Create a complete copy of the xlist (duplicate items, subitems,...).";

static char visit_doc[] = 
"L.visit([(xre0, function0),...(xreN, functionN)])\n\n\
Analise L and for each item that matchs xreI with I=0..N, call functionI, a Python function\n\
with a single parameter (the matching item).";


static PyMethodDef xlist_methods[] = {
	{"iter",		(PyCFunction)xlist_iter,		METH_VARARGS|METH_KEYWORDS, iter_doc		},
	{"match",		(PyCFunction)xlist_match,		METH_O,						match_doc		},
	{"validate",	(PyCFunction)xlist_validate,	METH_NOARGS,				validate_doc	},
	{"next_item",	(PyCFunction)xlist_next_item,	METH_NOARGS,				next_item_doc	},
	{"copy",		(PyCFunction)xlist_copy,		METH_NOARGS,				copy_doc		},
	{"query",		(PyCFunction)xlist_query,		METH_VARARGS|METH_KEYWORDS, query_doc		}, 
	{"visit",		(PyCFunction)xlist_visit,		METH_VARARGS,				visit_doc		}, 
	{NULL,	NULL},
};

static char xlist_doc[] =
"xlist() -> new xlist\n"
"xlist(args,kwds) -> new list with attributes from kwds and content from args";

PyTypeObject PyXList_Type = {
	PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,
	"xlist",
	sizeof(PyXListObject),
	0,
	(destructor)xlist_dealloc,					/* tp_dealloc */
	(printfunc) xlist_print,					/* tp_print */
	0,											/* tp_getattr */
	0,											/* tp_setattr */
	0,											/* tp_compare */
	(reprfunc)xlist_repr,						/* tp_repr */
	0,											/* tp_as_number */
	0,											/* tp_as_sequence */
	&xlist_as_mapping,							/* tp_as_mapping */
	0,											/* tp_hash */
	0,											/* tp_call */
	0,											/* tp_str */
	(getattrofunc)xlist_getattro,				/* tp_getattro */
	(setattrofunc)xlist_setattro,				/* tp_setattro */
	0,											/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE 
		| Py_TPFLAGS_HAVE_GC,					/* tp_flags */
	xlist_doc,									/* tp_doc */
 	(traverseproc)xlist_traverse,				/* tp_traverse */
 	(inquiry)xlist_clear,						/* tp_clear */
	(richcmpfunc)xlist_richcompare,				/* tp_richcompare */
	0,											/* tp_weaklistoffset */
	(getiterfunc)xlist_getiter,					/* tp_iter */
	0,											/* tp_iternext */
	xlist_methods,								/* tp_methods */
	0,											/* tp_members */
	0,											/* tp_getset */
	DEFERRED_ADDRESS(&PyList_Type),				/* tp_base */
	0,											/* tp_dict */
	0,											/* tp_descr_get */
	0,											/* tp_descr_set */
	0,											/* tp_dictoffset */
	(initproc)xlist_init,						/* tp_init */
	0, //PyType_GenericAlloc,						/* tp_alloc */
	xlist_new,									/* tp_new */
	0,//_PyObject_GC_Del,							/* tp_free */
};

void initxlist() {
	PyObject* __prefix__ = PyString_FromString("satine");

	PyObject* __attrs__ = Py_None;
	PyObject* __items__ = Py_None;
	__tag__ = PyString_FromString("__tag__");
	__uri__ = PyString_FromString("__uri__");
	__tpinfo__ = PyString_FromString("__tpinfo__");
	__dict__ = PyString_FromString("__dict__");

	PyXList_Type.tp_dict = PyDict_New();
	PyXList_Type.tp_base = &PyList_Type;
	if (PyType_Ready(&PyXList_Type) < 0)
		return;

	xlist_tpinfo = (PyXListInfo*) PyXListInfo_ForXList();
	PyDict_SetItemString(PyXList_Type.tp_dict, "tpinfo", (PyObject*) xlist_tpinfo);
}

