__version__ = "0.9"

__all__ = ["SoapServer"]

import os
import sys
import urllib
import urlparse
from BaseHTTPServer import HTTPServer
import SimpleHTTPServer
from satine.grapes.dispatcher import Dispatcher
from satine.grapes import grape
from satine import kernel
import satine.schemas.soap
import satine

import shutil

class Form(satine.xlist):
    __etable__ = []

class SoapServerHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
   def do_SOAP(self):
      path,query = self.__get_info()
      length = int(self.headers.getheader('content-length'))

      data = self.rfile.read(length)
      soap_message = satine.xml2py(data)
      soap_body = soap_message.iter(".<soap:Body>|").next()
      if soap_body:
          message = soap_body[0]
          response, code, mimetype = self.server.dispatcher.process(path, message, 200, "text/xml")
          print response

      envelope = satine.xml2py("<soap:Envelope><soap:Body/></soap:Envelope>")
      envelope[0].append(response)
      response = satine.py2xml(envelope)
      self.send_response(code)
      self.send_header("Content-type", mimetype)
      self.send_header("Content-length", str(len(response)))
      self.end_headers()
      self.wfile.write(response);
      
   def do_POST(self):
      """Serve a POST request """
      if self.headers.typeheader is None:
         content_type = self.headers.type
      else:
         content_type = self.headers.typeheader

      if content_type == "application/soap+xml":
         return self.do_SOAP()

      dir,name,query = self.__get_info()
      length = int(self.headers.getheader('content-length'))

      data = self.rfile.read(length)
      message = self.__get_form(data)
      response, code, mimetype = self.server.dispatcher.process(path, message, 200, "text/xml")
      if satine.is_xlist(response):
          response = """<?xml version="1.0"?>"""+satine.py2xml(message)

      self.send_response(code)
      self.send_header("Content-type", mimetype)
      self.send_header("Content-length", str(len(response)))
      self.end_headers()
      self.wfile.write(response);
      
   def do_GET(self):
      """Serve a GET request """
      print "GET request", self.path
      path,query = self.__get_info()
       
      message = self.__get_form(query)
      response, code, mimetype = self.server.dispatcher.process(path, message, 200, "text/xml")
      if satine.is_xlist(response):
          response = """<?xml version="1.0"?>"""+satine.py2xml(message)

      self.send_response(code)
      self.send_header("Content-type", mimetype)
      self.send_header("Content-length", str(len(response)))
      self.end_headers()
      self.wfile.write(response);

   def __dispatch(self, message):
      for disp in self.dispatchers:
         disp.receive(message)
            
   def __get_info(self):
      protocol,host,path,params,query,fragment = urlparse.urlparse(self.path)
      return path,query

   def __get_form(self, query):
      form = Form()
      query = query.replace("\n","").replace('+', ' ')
      vars = query.split("&")
      for var in vars:
         if var:
            name, value = var.split("=")
            form[name] = value
      return form


class Echo(grape):
    def process(self, message, code, mimetype):
        return message, code, mimetype
    
class Xml2Html(grape):
    def process(self, message, code, mimetype):
        doc = """<?xml version="1.0"?>\n"""+satine.py2xml(message)
        doc = doc.replace("&","&amp;").replace("<","&lt;").replace(">","&gt;").replace('"',"&quot;")
        doc = doc.replace("&lt;","<b>&lt;</b>").replace("&gt;","<b>&gt;</b>").replace("\n","<br>")
        page = """<html><body>"""+doc+"""
                    <br><hr>xml to html converter (from satine.HTTP.Xml2Html)</body></html>"""
        return page, code, "text/html"

class FileLoader(grape):
    def __init__(self, sender, address, model, wwwroot):
        grape.__init__(self, sender, address, model)
        self.fspath = wwwroot+address

    def process(self, message, code, mimetype):
        try:
            fd = open(self.fspath, "rb")
            doc = fd.read()
            head,ext = os.path.splitext(self.fspath)
            ext = ext.lower()
            if ext == '.gif':
                mimetype = "image/gif" 
            elif ext == '.jpg':
                mimetype = "image/jpeg" 
            elif ext == '.jpeg':
                mimetype = "image/jpeg" 
            elif ext == '.html':
                mimetype = "text/html" 
            elif ext == '.htm':
                mimetype = "text/html" 
            elif ext == '.txt':
                mimetype = "text/plain" 
        except:
            doc = """
                <html>
                <head>
                  <title>Error Message from Satine HtmlFilter</title>
                </head>
                <body>
                    <h1>Hei!!!Error</h1><br><br>
                    I am the Satine HtmlFilter. My task is to translate a XML message to a
                    HTML page. <br>
                    Unfortunately, the resource you requested is not available.<br>

                    - filename: <b>%s </b><br>
                    - document root: <b> %s </b><br>
                    - HTTP path: <b>%s</b><br>                    
                </body>
                </html>""" % (self.fspath, wwwroot, address)
            mimetype = "text/html"
        return doc, code, mimetype

class HtmlFilter(grape):
    def __init__(self, sender, address, model, wwwroot):
        grape.__init__(self, sender, address, model)
        self.fspath = wwwroot+address
        self.wwwroot = wwwroot
        self.address = address
        self.validfile = 1

    def load(self, message, code, mimetype):
        try:
            fd = open(self.fspath, "rb")
            self.doc = fd.read()
        except:
            original = message
            if mimetype == "text/xml":
                original = """<?xml version="1.0"?>\n"""+satine.py2xml(message)
                original = original.replace("&","&amp;").replace("<","&lt;").replace(">","&gt;").replace('"',"&quot;")
                original = original.replace("&lt;","<b>&lt;</b>").replace("&gt;","<b>&gt;</b>").replace("\n","<br>")
            self.doc = """
                <html>
                <head>
                  <title>Error Message from Satine HtmlFilter</title>
                </head>
                <body>
                    <h1>Hei!!!Error</h1><br><br>
                    I am the Satine HtmlFilter. My task is to translate a XML message to a
                    HTML page. <br>
                    Unfortunately, the Satine Service Page you requested is not available.<br>

                    - filename: <b>%s </b><br>
                    - document root: <b> %s </b><br>
                    - HTTP path: <b>%s</b><br>

                    The original message was:<br>
                    %s                                        
                </body>
                </html>""" % (self.fspath, self.wwwroot, self.address, original)

        self.commands = []
        doc = self.doc
        idx = doc.find("@")
        while idx != -1:
            start = idx
            idx = doc.find("@", idx+1)
            pattern = doc[start+1:idx]
            self.commands.append((pattern, start, idx))
            if (idx != -1):
                idx = doc.find("@", idx+1)            
                
    def process(self, message, code, mimetype):
        res = ""
        pos = 0
        buffer = []
        loop = 0
        cache = {}
        self.load(message, code, mimetype)
        
        for cmd, left, right in self.commands:
            if cmd == "begin":
                loop = 1
                res += self.doc[pos:left]
                pos = right+1
            elif cmd == "end":
                i = 0
                last = 0
                bg = pos
                while (not last) and len(buffer):
                    for vals, l, r in buffer:
                        if len(vals) == i+1:
                            last = 1
                        res += self.doc[pos:l]
                        if len(vals[i]):
                            res += vals[i]
                        else:
                            res += "None"
                        pos = r+1
                    res += self.doc[pos:left]
                    i += 1
                    pos = bg
                loop = 0
                pos = right+1
            elif cmd == "":
                res += "@"
            else:
                try:
                    if cmd in cache:
                        vals = cache[cmd]
                    else:
                        vals = message.query(cmd)
                        cache[cmd] = vals
                    if not loop:
                        res += self.doc[pos:left]
                        if len(vals):
                            res += vals[0]
                        else:
                            res += "None"
                        pos = right+1
                    else:
                        if len(vals):
                            buffer.append((vals,left,right))
                except:
                    pass

        res += self.doc[pos:]
        return res, code, "text/html"
                
def run(dispatcher, port = 7580, HandlerClass = SoapServerHandler, ServerClass = HTTPServer):
   server_address = ('', port)
   httpd = ServerClass(server_address, HandlerClass)
   httpd.dispatcher = dispatcher
   
   sa = httpd.socket.getsockname()
   print "Serving Satine WS server on", sa[0], "port", sa[1], "..."
   httpd.serve_forever()
      
def test():
   dispatcher = Dispatcher()
   dispatcher.load("HTTP.test.config.xml")
   run(dispatcher)

satine.xspace(satine="http://satine.sourceforge.net/schemas/kernel")

if __name__ == '__main__':
    test()