import getopt, sys
from satine import xml2py, dt
import satine.schemas.xmlschema
import time
import StringIO
import keyword
import satine

#satine.verbose(2)
class schemac:
    def compile(self, document, prefix, uri = None):
        self.document = document
        self.prefix = prefix
        self.elements = []
        self.reduce()
        if self.schema.targetNamespace:
            self.uri = self.schema.targetNamespace
        else:
            self.uri = uri

        return self.generate()

    def reduce(self):
        self.schema = xml2py(self.document)
        error = self.schema.validate()
        if error:
            raise str(error)
        self.reduceSchema(self.schema)

    def generate(self):
        pass

    def reduceSchema(self, e):
        for el in e.query(".<xsd:element>|"):
            self.reduceElement(el)
            
        for ct in e.query(".<xsd:complexType>|"):
            self.reduceComplexType(ct)        
            
        for ct in e.query(".<xsd:simpleType>|"):
            self.reduceSimpleType(ct)        

    def reduceElement(self, e):
        if e.name == None and e.ref == None:
            raise "invalid element declaration: missing element name"
        if e.ref:
            return
                        
        imptype = None
        for ct in e.query(".<xsd:complexType>|"):
            self.reduceComplexType(ct)
            imptype = ct.name
            e.remove(ct)
            self.schema.append(ct)
            
        for st in e.query(".<xsd:simpleType>|"):
            self.reduceSimpleType(st)
            imptype = st.name
            e.remove(st)
            self.schema.append(st)

        if e.type and imptype:
            raise "double type for element definition %s" % e
        elif not e.type and not imptype:
            raise "none type for element definition %s" % e
        elif not e.type and imptype:
            e.type = imptype
        
    def reduceComplexType(self, e):
        if not e.name:
            e.name = 'anonymousType_%f' % (time.time(),)
            e.name = e.name.replace('.','_')

        for el in e.query(".<xsd:choice>|"):
            self.reduceContainer(el)
        for el in e.query(".<xsd:sequence>|"):
            self.reduceContainer(el)

    def reduceSimpleType(self, e):
        if not e.name:
            e.name = 'anonymousType_%f' % (time.time(),)
            e.name = e.name.replace('.','_')

        for el in e.query(".<xsd:list>|"):
            self.reduceList(el)
        for el in e.query(".<xsd:union>|"):
            self.reduceUnion(el)

    def reduceList(self, e):
        imptype = None

        for st in e.query(".<xsd:simpleType>|"):
            self.reduceSimpleType(st)
            imptype = st.name
            e.remove(st)
            self.schema.append(st)

        if e.itemType and imptype:
            raise "double itemType for list definition %s" % e
        elif not e.itemType and not imptype:
            raise "none itemType for list definition %s" % e
        elif not e.itemType and imptype:
            e.itemType = imptype

    def reduceUnion(self, e):
        imptype = ""

        for st in e.query(".<xsd:simpleType>|"):
            pass

        if e.type and imptype:
            raise "double itemType for list definition %s" % e
        elif not e.type and not imptype:
            raise "none itemType for list definition %s" % e
        elif not e.type and imptype:
            e.itemType = imptype

    def reduceContainer(self, e):
        for el in e.query(".<xsd:element>|"):
            self.reduceElement(el)
        for el in e.query(".<xsd:choice>|"):
            self.reduceContainer(el)
        for el in e.query(".<xsd:sequence>|"):
            self.reduceContainer(el)
                       

class pythonc(schemac):
    def generate(self):
        self.code = """
#Satine schema compiler  2-20-2003
import satine
"""
        if self.uri == None:
            raise "undefined target namespace"

        self.buildComplexTypes()
        self.buildSimpleTypes()
        self.buildElements()
        self.code += """\n\nsatine.xspace(%s = "%s")""" % (self.prefix,self.uri)
        return self.code

    def buildComplexTypes(self):
        self.types = {}
        for ct in self.schema.query(".<xsd:complexType>|"):
            self.types[ct.name] = self.buildComplexType(ct)

    def buildSimpleTypes(self):
        self.simpleTypes = {}
        for st in self.schema.query(".<xsd:simpleType>|"):
            self.simpleTypes[st.name] = st

        self.simpleTypesList = []
        for st in self.simpleTypes.values():
            self.orderSimpleType(st)
            
        for name in self.simpleTypesList:
            self.code += self.buildSimpleType(self.simpleTypes[name])

    def buildComplexType(self, ct, base=None):
        __attrs__ = self.decodeAttrs(ct)
        complexContent = xml2py("<xsd:complexContent/>")
        simpleContent = xml2py("<xsd:simpleContent/>")
        for item in ct:
            if item.match(complexContent):
                return  self.buildComplexContent(item)
            elif item.match(simpleContent):
                return self.buildSimpleContent(item)
        __items__ = self.decodeItems(ct)
        return (base, __attrs__, __items__, None)

    def decodeItems(self, ct):
        sequence = xml2py("<xsd:sequence/>")
        choice = xml2py("<xsd:choice/>")
        all = xml2py("<xsd:all/>")

        __items__ = None        
        for item in ct:
            if item.match(sequence):
                __items__ = "'%s'" % self.decodeSequence(item)
            elif item.match(choice):
                __items__ = "'%s'" % self.decodeChoice(item)
            elif item.match(all):
                __items__ = "'%s'" % self.decodeAll(item)
        return __items__

    def buildComplexContent(self, cc):
        restriction = xml2py("<xsd:restriction/>")
        extension = xml2py("<xsd:extension/>")

        for item in cc:
            if item.match(extension):
                res = self.buildExtension(item)
            elif item.match(restriction):
                res = self.buildExtension(item)
        return res

    def buildExtension(self, et):
        if not et.base :
            raise "extension without basetype"

        prefix, tag =  dt.QName.decode(et.base)
        if prefix:
            cls = "satine.get_type(%s,%s)"%(tag,prefix)
        elif prefix == 'xsd':
            cls = "dt."+tag
        else:
            cls = et.base

        return self.buildComplexType(et, cls)

    def buildSimpleTypeRestriction(self, rst):
        __facets__ = ""
        if rst.base == None:
            raise "restriction needs a base type"        

        for item in rst:
            __facets__ += str(item)[:-2]+">\n\t\t"
        if __facets__:
            return (rst.base, '"""\n\t\t%s"""' % __facets__)
        else:
            return (rst.base, None)
        
    def buildSimpleType(self, st):
        restriction = xml2py("<xsd:restriction/>")
        _list = xml2py("<xsd:list/>")
        union = xml2py("<xsd:union/>")

        for item in st:
            if item.match(restriction):
                res = self.buildSimpleTypeRestriction(item)
            elif item.match(_list):
                res = self.buildList(item)
            elif item.match(_list):
                res = self.buildUnion(item)

        name = st.name
        base, facets = res

        while keyword.iskeyword(name):
            name = '_'+name

        code = ""
        prefix, tag = dt.QName.decode(base)
        if prefix:        
            code += "%s_parent = satine.getclass('%s')" % (name, base)
        else:
            code += "%s_parent = %s" % (name, base)
                    
        code += "\nclass %s(%s_parent):\n" % (name, name)
        code += "\n\t__tag__ = '%s'\n" % (st.name)

        if facets:
            code += "\t__facets__=%s_parent.__facets__+%s\n\n" % (name, facets)
        else:
            code += "\tpass\n\n"
        return code
      
    def orderSimpleType(self, st):
        if st.name in self.simpleTypesList:
            return
        
        restriction = xml2py("<xsd:restriction/>")
        _list = xml2py("<xsd:list/>")
        union = xml2py("<xsd:union/>")

        for item in st:
            if item.match(restriction):
                base = item.base
            elif item.match(_list):
                base = item.itemType
            elif item.match(union):
                base = item.itemType

        if base in self.simpleTypes:
            self.orderSimpleType(self.simpleTypes[base])
        self.simpleTypesList.append(st.name)
      

    def buildList(self, ls):
        if ls.itemType == None:
            raise "list needs an item type"        

        prefix, tag =  dt.QName.decode(ls.itemType)
        if prefix:
            cls = '"""<%s>"""' %ls.itemType
        else:
            cls = '"""<%s:%s>"""' % (self.prefix,ls.itemType)
        return ("xsd:dtlist", cls)
        

    def buildElements(self):
        for el in self.schema.query(".<xsd:element>|"):
            self.code += self.buildElement(el)

        for el in self.elements:        
            self.code += self.buildElement(el)

    def subTypeFill(self, type):
        if type not in self.types:
            return None, None, None, None
        subType, __attrs__, __items__, __facets__ = self.types[type]
        if not subType:
            return subType,  __attrs__, __items__, __facets__

        nbt, n_attrs, n_items, n_facets = self.subTypeFill(subType)
        if n_attrs:
            if __attrs__:
                __attrs__ = __attrs__[:-1] + n_attrs[1:]
            else:
                __attrs__ = n_attrs
        if n_items:
            if __items__:
                __items__ = __items__[:-1] + n_items[1:]
            else:
                __items__ = n_items
        return subType,  __attrs__, __items__, __facets__
            

    def buildElement(self, el):
        name = el.name
        while keyword.iskeyword(name):
            name = '_'+name

        baseType, __attrs__, __items__, __facets__ = self.subTypeFill(el.type)
            
        code = "\nclass %s(satine.xlist):\n" % (name)
        if name != el.name:
            code += "\t__tag__='%s'\n" % el.name

        if __attrs__:
            code += "\t__attrs__=%s\n" % __attrs__
        if __items__:
            code += "\t__items__=%s\n" % __items__
        if __facets__:
            code += "\t__facets__=%s\n" % __facets__
        if not __attrs__ and not __items__ and not __facets__:
            code += "\tpass\n"
        if __facets__:
            code += "%s.compile_facets()" % name
        return code

    def decodeSequence(self, sq):
        __items__ = ""
        element = xml2py("<xsd:element/>")
        any = xml2py("<xsd:any/>")
        sequence = xml2py("<xsd:sequence/>")
        choice = xml2py("<xsd:choice/>")
        all = xml2py("<xsd:all/>")

        for item in sq:
            if item.match(element):
                if item.type:
                    self.elements.append(item)
                if item.name:
                    __items__ += "<%s>" % item.name
                elif item.ref:
                    __items__ += "<%s>" % item.ref
                else:
                    raise "element without ref and name"
                if item.maxOccurs > 1:
                    __items__ += "*"
                elif item.minOccurs == 0:
                    __items__ += "?"
            elif item.match(any):
                __items__ += "."
            elif item.match(sequence):
                __items__ += "(%s)" % self.decodeSequence(item)
            elif item.match(choice):
                __items__ += "(%s)" % self.decodeChoice(item)
            elif item.match(all):
                __items__ = "(%s)" % self.decodeAll(item)
            else:
                raise "item %s unattended" % item
        return __items__
            
    def decodeAll(self, sq):
        return "(%s)*" % self.decodeChoice(sq)

    def decodeChoice(self, sq):
        __items__ = ""
        element = xml2py("<xsd:element/>")
        any = xml2py("<xsd:any/>")
        sequence = xml2py("<xsd:sequence/>")
        choice = xml2py("<xsd:choice/>")
        all = xml2py("<xsd:all/>")

        for item in sq:
            if item.match(element):
                if item.type:
                    self.elements.append(item)
                if __items__:
                    __items__ += "|"
                if item.name:
                    __items__ += "<%s>" % item.name
                elif item.ref:
                    __items__ += "<%s>" % item.ref
                else:
                    raise "element without ref and name"
                if item.maxOccurs > 1:
                    __items__ += "*"
                elif item.minOccurs == 0:
                    __items__ += "?"
            elif item.match(any):
                __items__ += "."
            elif item.match(sequence):
                __items__ += "(%s)" % self.decodeSequence(item)
            elif item.match(choice):
                __items__ += "(%s)" % self.decodeChoice(item)
            elif item.match(all):
                __items__ = "(%s)" % self.decodeAll(item)
            else:
                raise "item %s unattended" % item
        return __items__
            
    def decodeAttrs(self, ct):
        __attrs__ = ""
        for attr in ct.query(".<xsd:attribute>|"):
            if not attr.name:
                raise "No name in attribute definition %s" % attr
            if not attr.type:
                attr.type = "xsd:string"

            if attr.type.count(':'):
                __attrs__ += "<%s" % attr.type
            else:
                __attrs__ += "<%s:%s" % (self.prefix,attr.type)
                
            if attr.default:
                __attrs__ += ' default="%s"'+attr.default
            if attr.use == 'required':
                __attrs__ += ' use="required"'
            __attrs__ += ">%s" % attr.name

        if __attrs__:
            return "'%s'" % __attrs__
        else:
            return None

class htmlc(schemac):
    def generate(self):
        if self.targetNamespace == None:
            raise schemacError("undefined target namespace")

        self.code = """
                    <html>
                    <head>
                    <title>Satine Schema Description</title>
                    <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
                    </head>

                    <body bgcolor="#FFFFFF" text="#000000">
                    <div align="center"> 
                      <font size="+2"><b>Satine Schema Description</b></font><br>
                      <font size="+2"></font><i>"""+self.targetNamespace+"""</i>
                    </div>

                    <br><br>""" 
        self.defines()
        self.code += """
                    <div align="center"> 
                    <br><i>compiled on """+ time.asctime()+"""</i> 
                    </div>
                    </body>
                    </html>""" 
        return self.code

    def defines(self):
        for name,tp in self.elements:
            if name.startswith('@'):
                continue

            mt,tn = self.getType(tp)
            if mt == 'native':
                self.code += "<b>"+name+"</b>: "+datatypes.dQName().py2xml(tp)+"<br>\n<hr>\n"
            elif mt == 'simple':
                typ = self.types[tn]
                if typ.base == None:
                    raise SchemacError("simple type "+name+" has not a base class")                                        
                bt, base = self.getType(typ.base)
                if self.extensibles.count(name):
                    self.code += "<b><i>"+name+"</i></b>: "+base+"<br>\n"
                else:
                    self.code += "<b>"+name+"</b><br>\n"
                typ = self.types[tn]
                for i in typ.restrictions.keys():
                    if (i == 'enumeration'):
                        for j in typ.restrictions[i]:
                            self.code += "<i>-enumeration = "+str(j)+"</i><br>\n"
                    else:
                        self.code += "<i>-"+i+" = "+str(typ.restrictions[i])+"</i></br>\n"
                self.code += "<hr>\n"
            elif mt == 'complex':
                typ = self.types[tn]
                self.code += "<b>"+name+"</b><br>\n"
                for aname,atype,adef in typ.atable:
                    if adef:
                        self.code += "<li type=\"circle\">"+aname+":"+atype+" = "+adef+"</li>\n"
                    else:
                        self.code += "<li type=\"circle\">"+aname+": "+datatypes.dQName().py2xml(atype)+"</li>\n"
                for ename,etype in typ.etable:
                    self.code += "<li type=\"square\">"+ename+"</li>\n"
                self.code += "<hr>\n"
            else:
                raise schemacError("unknown metatype")


def loadXMLSchema(filename):
    file = open(filename)
    doc = file.read()
    code = pythonc().compile(doc,"time", "hdfskj")
    print code

if __name__ == '__main__':
    loadXMLSchema("2511B39.xsd")
    

    