import satine
import unittest
import gc
import satine.schemas.soap
import satine.test
import os
from satine import schemaclib

def openfile(filename):
    testpath = satine.test.__path__[0]
    filename = testpath+os.sep+filename
    return open(filename)

class xlistFeatures(unittest.TestCase):
    def testTpinfo(self):
        l = satine.xlist()
        self.assertEqual(l.tpinfo.tag, "xlist")
        self.assertEqual(l.tpinfo.uri, "http://satine.sourceforge.net/schemas/kernel")
        self.assertEqual(l.tpinfo.prefix, "satine")
        self.assertEqual(l.tpinfo.attrs_types, None)
        self.assertEqual(l.tpinfo.default_attrs, {})
        self.assertEqual(l.tpinfo.items_dfa, None)
        self.assertEqual(l.tpinfo.none_attrs, [])
        self.assertEqual(l.tpinfo.required_attrs, [])
       
    def testConstructor(self):
        l = satine.xlist("first content", "second content", attr1="value1", attr2="value2")
        self.assertEqual(l[0], "first content")
        self.assertEqual(l[1], "second content")
        self.assertEqual(l.attr1, "value1")
        self.assertEqual(l.attr2, "value2")

    def testAppend(self):
        l = satine.xlist()
        l.append("first content")
        l.append("second content")
        self.assertEqual(l[0], "first content")
        self.assertEqual(l[1], "second content")
        
    def testInsert(self):
        l = satine.xlist()
        l.insert(0, "second content")
        l.insert(0, "first content")
        self.assertEqual(l[0], "first content")
        self.assertEqual(l[1], "second content")
        
    def testRemove(self):
        l = satine.xlist("first content", "second content")
        self.assertEqual(l[0], "first content")
        l.remove("first content")
        self.assertEqual(l[0], "second content")
        del l[0]
        self.assert_(len(l) == 0)

    def testCount(self):
        l = satine.xlist("first content", "second content", "second content")
        self.assertEqual(l.count("first content"), 1)
        self.assertEqual(l.count("second content"), 2)

    def testIndex(self):
        l = satine.xlist("first content", "second content", "second content")
        self.assertEqual(l.index("first content"), 0)
        self.assertEqual(l.index("second content"), 1)
        self.assertRaises(ValueError, l.index, "item not inside")

    def testReverse(self):
        l = satine.xlist("second content", "first content")
        l.reverse()
        self.assertEqual(l[0], "first content")
        self.assertEqual(l[1], "second content")

    def testSort(self):
        l = satine.xlist('b','a','z','g',satine.xlist(0), 'c','j','k','d','u',2,32,1,59,0,-21)
        l.sort()
        ordered =[-21,0,1,2,32,59,satine.xlist(0),'a','b','c','d','g','j','k','u','z']
        for i in range(len(l)):
            self.assertEqual(l[i], ordered[i])

    def testSortXLists(self):
        i = satine.xlist(attr1="1", attr2="1")
        j = satine.xlist(attr1="2", attr2="1")
        k = satine.xlist(attr1="2", attr2="2")
        l = satine.xlist(j, i, k)
        self.assertEqual(l[0], j)
        self.assertEqual(l[1], i)
        self.assertEqual(l[2], k)
        l.sort()
        self.assertEqual(l[0], i)
        self.assertEqual(l[1], j)
        self.assertEqual(l[2], k)
        
    def testPop(self):
        l = satine.xlist("first content", "second content")
        self.assertEqual(l.pop(), "second content")
        self.assertEqual(l.pop(), "first content")

    def testMatch(self):
        l = satine.xlist("first content", "second content", attr1="value1", attr2="value2")
        k = satine.xlist("first content", attr1="value1", attr2="value2")
        self.assert_(l.match(k))
        k.attr1 = "new value"
        self.assert_(l.match(k) == 0)
        del k["attr1"]
        del k["attr2"]
        self.assert_(l.match(k))
        self.assert_(k.match(l) == 0)

    def testCopy(self):
        l = satine.xlist("first content", "second content", attr1="value1", attr2="value2")
        cp = l.copy()
        self.assertEqual(l, cp)
        

class Conversion(unittest.TestCase):
    def testSoapMessage(self):
        doc = openfile('simplesoap2.xml').read()
        py = satine.xml2py(doc)
        xml = satine.py2xml(py)
        self.assertEqual(xml, doc)
        self.assertEqual(py.tpinfo.uri, "http://schemas.xmlsoap.org/soap/envelope/")

class Query(unittest.TestCase):
    def setUp(self):
        doc = openfile('simplesoap1.xml').read()
        self.data = satine.xml2py(doc)

    def testTagQuery(self):
        res = self.data.query("Body",style="tag")
        self.assertEqual(res, satine.xlist(self.data[1]))
        
    def testXREQuery(self):
        res = self.data.query("<soap:Envelope><soap:Body>|")
        self.assertEqual(res, satine.xlist(self.data[1]))
        res = self.data.query(".<soap:Body>|")
        self.assertEqual(res, satine.xlist(self.data[1]))
        res = self.data.query("<soap:Envelope>*<soap:Body>|")
        self.assertEqual(res, satine.xlist(self.data[1]))
        res = self.data.query("<soap:Envelope><soap:Body>.|")
        self.assertEqual(res, satine.xlist(self.data[1][0]))
        res = self.data.query("<soap:Body>|")
        self.assert_(res != satine.xlist(self.data[1][0]))
        self.assert_(res == satine.xlist())

    def testExtractionQuery(self):
        res = self.data.query("<soap:Envelope><soap:Body>..|")
        self.assertEqual(res, satine.xlist("this is a test"))
        res = self.data.query("<soap:Envelope><soap:Body>.|from")
        self.assertEqual(res, satine.xlist("unittest"))

class Visit(unittest.TestCase):
    def setUp(self):
        doc = openfile('zoo.xml').read()
        self.py = satine.xml2py(doc)

    def tiger(self, e):
        self.names.append(e.name)
        for diet in e:
            self.food.append(diet[0])
        
    def monkey(self, e):
        self.names.append(e.name)
        for diet in e:
            self.food.append(diet[0])
        self.assert_(hasattr(e, "age"))
        self.assertEqual(e.age, "14")

    def giraffe(self, e):
        self.names.append(e.name)
        for diet in e:
            self.food.append(diet[0])
        
    def testVisit(self):
        self.names = []
        self.food = []
        self.py.visit([
            ('<zoo><tiger>',self.tiger),
            ('<zoo><monkey name="ugo">',self.monkey),
            ('<zoo><giraffe>',self.giraffe)
        ])

        self.assertEqual(self.names, ['leo', 'ugo', 'anita'])
        self.assertEqual(self.food, ['beefsteak', 'chicken', 'banana', 'peanut', 'corn', 'hay'])

class Validation(unittest.TestCase):
    def testSoapItems(self):
        py = satine.xml2py("<soap:Envelope/>")
        self.assertEqual(py.validate(), [0, 'not enough items'])
        py = satine.xml2py("<soap:Envelope><soap:Body>hello world</soap:Body></soap:Envelope>")
        self.assertEqual(py.validate(), [])        
        py = satine.xml2py("<soap:Envelope><soap:Header></soap:Header></soap:Envelope>")
        self.assertEqual(py.validate(), [1, 'not enough items'])        
        py = satine.xml2py("<soap:Envelope>hello world<soap:Body/></soap:Envelope>")
        self.assertEqual(py.validate(), [0, "error at position 0: found 'hello world' while one of [<soap:Header/>, <soap:Body/>] needed"])        
        py = satine.xml2py("<soap:Envelope><soap:Body/><soap:Header/></soap:Envelope>")
        self.assertEqual(py.validate(), [1, 'error at position 1: found <soap:Header/> while one of [None] needed'])        
    
class ComplexAttrs(unittest.TestCase):
    def setUp(self):
        self.py = satine.xlist()
        self.py["soap:attr"] = "soap_attr"
        self.py["attr"] = "attr"

    def testGet(self):
        self.assertEqual(self.py["soap:attr"], "soap_attr")
        self.assertEqual(self.py["attr"], "attr")

    def _testPrefix(self):
        self.py = satine.xml2py('<satine:xlist xmlns:xlisttest="http://satine.sourceforge.net/temp/xlisttest" soap:attr="soap" xlisttest:attr="test"/>')
        doc = satine.py2xml(self.py, {"http://schemas.xmlsoap.org/soap/envelope/":"SOAP",
                                      "http://satine.sourceforge.net/schemas/kernel":""})
        self.assertEqual(doc, '<xlist xmlns="http://satine.sourceforge.net/schemas/kernel" xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/" SOAP:attr="soap" xmlns:xlisttest="http://satine.sourceforge.net/temp/xlisttest" xlisttest:attr="test"/>')


if __name__ == "__main__":
    unittest.main()